package miniteller;

import java.sql.*;
import java.util.Vector;

public class MiniTellerServ extends _MiniTellerImplBase {

  String DRIVER = System.getProperty("MiniTeller.driver");
  String DBURL = System.getProperty("MiniTeller.dburl");
  String USERNAME = System.getProperty("MiniTeller.username");
  String PASSWORD = System.getProperty("MiniTeller.password");

  static final String STMT_INSERT_NEW =
    "INSERT INTO ACCOUNTS (ACCTNO, BALANCE) VALUES (?, ?)";
  static final String STMT_GET_ACCTNO = "SELECT SEQ_ACCTNO.NEXTVAL FROM DUAL";
  static final String STMT_SET_NAME =
    "UPDATE ACCOUNTS SET FIRSTNAME = ?, LASTNAME = ? WHERE ACCTNO = ?";
  static final String STMT_GET_LIST =
    "SELECT ACCTNO, FIRSTNAME, LASTNAME, BALANCE FROM ACCOUNTS " +
    "ORDER BY LASTNAME, FIRSTNAME";
  static final String STMT_GET_ACCTINFO =
    "SELECT ADDR1, ADDR2, PHONE, SSN FROM ACCTINFO WHERE ACCTNO = ?";
  static final String STMT_SET_ACCTINFO =
    "INSERT INTO ACCTINFO (ACCTNO, ADDR1, ADDR2, PHONE, SSN) " +
    "VALUES (?, ?, ?, ?, ?)";
  static final String STMT_GET_BAL =
    "SELECT BALANCE FROM ACCOUNTS WHERE ACCTNO = ?";
  static final String STMT_SET_BAL =
    "UPDATE ACCOUNTS SET BALANCE = ? WHERE ACCTNO = ?";

	// approximate maximum storable in DB column
	protected static final float AMT_MAX = 10000000;

  protected AccountList l;

  protected Connection c = null;
  protected Statement s = null;
  protected PreparedStatement psNew, psAcctNo, psSetName, psGetList,
    psGetInfo, psSetInfo, psGetBal, psSetBal;

  protected boolean valid = false;

  public MiniTellerServ(String name) {
    super(name);
    valid = dbinit();
  }

  public MiniTellerServ() {
    super();
    valid = dbinit();
  }

  public void beginTransaction() throws AccessViolation {
    try {
    if (!c.getAutoCommit()) throw new AccessViolation(
     "Internal error: Failure of synchronization");
    System.out.println("Beginning transaction...");
    c.setAutoCommit(false);
    } catch (SQLException e) { throw new AccessViolation(e); }
  }

  public void commitTransaction() throws AccessViolation {
    try {
    if (c.getAutoCommit()) throw new AccessViolation(
     "Internal error: Can't commit when connection is set to auto-commit");
    c.commit();
    System.out.println("Committed transaction.");
    c.setAutoCommit(true);
    } catch (SQLException e) { throw new AccessViolation(e); }
  }

  public void rollbackTransaction() throws AccessViolation {
    try {
    if (!c.getAutoCommit()) throw new AccessViolation(
     "Internal error: Can't rollback when connection is set to auto-commit");
    c.rollback();
    System.out.println("Transaction rolled back.");
    c.setAutoCommit(true);
    } catch (SQLException e) { throw new AccessViolation(e); }
  }

  // internal methods called by AccountServ class for access to database

  float getBal(Account acct) throws AccessViolation {
    ResultSet r = null;
    try {
      psGetBal.setInt(1, acct.acctno());
      r = psGetBal.executeQuery();
      r.next();
      return r.getFloat(1);
    } catch (SQLException e) {
      throw new AccessViolation(
        "Database error while retrieving balance: " + e.getMessage());
    } finally {
      try { r.close(); } catch (Exception e) {}
    }
  }

  void setBal(Account acct, float bal) throws AccessViolation {
		if (bal > AMT_MAX) throw new AccessViolation(
			"Amount too large for database");
    try {
      psSetBal.setInt(1, acct.acctno());
      psSetBal.setFloat(2, bal);
      psSetBal.executeUpdate();
    } catch (SQLException e) {
      throw new AccessViolation(
        "Database error while setting balance: " + e.getMessage());
    }
  }

  void setCustInfo(AccountServ acct, CustInfo info) throws AccessViolation {
    int acctno = acct.acctno();
    try {
      psSetName.setString(1, acct.firstName);
      psSetName.setString(2, acct.lastName);
      psSetName.setInt(3, acctno);
      psSetInfo.setInt(1, acctno);
      psSetInfo.setString(2, info.addr1);
      psSetInfo.setString(3, info.addr2);
      psSetInfo.setString(4, info.phone);
      psSetInfo.setLong(5, info.ssn);
      psSetName.executeUpdate();
      psSetInfo.executeUpdate();
    } catch (SQLException e) {
      throw new AccessViolation(
        "Database error while setting account information: " + e.getMessage());
    }
  }
  
  CustInfo getCustInfo(Account acct) throws AccessViolation {
    ResultSet r = null;
    try {
      psGetInfo.setInt(1, acct.acctno());
      r = psGetInfo.executeQuery();
      if (!r.next()) return new CustInfo();
      return new CustInfo(r.getString(1), r.getString(2),
			  r.getString(3), r.getLong(4));
    } catch (SQLException e) {
      throw new AccessViolation(
        "Database error while getting account information: " + e.getMessage());
    } catch (DataInvalid de) { // ouch.
      System.err.println("CustInfo retrieved from database is invalid!\n" +
			 de.getMessage());
      return null;
    } finally {
      try { r.close(); } catch (Exception e) {}
    }
  }

  protected boolean dbinit() {
    System.out.println("Connecting to database...");
    try {
    Class.forName(DRIVER);
    } catch (ClassNotFoundException e) {
      System.err.println("Unable to load class: " + e.getMessage());
      e.printStackTrace();
      return false;
    }

    try {
    // connect
    c = DriverManager.getConnection(DBURL, USERNAME, PASSWORD);
    s = c.createStatement();
    // prepared statements
    psNew = c.prepareStatement(STMT_INSERT_NEW);
    psAcctNo = c.prepareStatement(STMT_GET_ACCTNO);
    psSetName = c.prepareStatement(STMT_SET_NAME);
    psGetList = c.prepareStatement(STMT_GET_LIST);
    psGetInfo = c.prepareStatement(STMT_GET_ACCTINFO);
    psSetInfo = c.prepareStatement(STMT_SET_ACCTINFO);
    psGetBal = c.prepareStatement(STMT_GET_BAL);
    psSetBal = c.prepareStatement(STMT_SET_BAL);

    // initialize account list
    l = new AccountList();
    readList();
    } catch (Exception e) {
      System.err.println("Unable to initialize database connection: " +
			 e);
      e.printStackTrace();
      if (e instanceof SQLException)
        System.err.println(((SQLException) e).getSQLState());
      return false;
    }

    return true;
  }

  protected void readList() throws SQLException {
    ResultSet r = psGetList.executeQuery();
    while (r.next())
      l.addAccount(
        new AccountServ(this, r.getInt(1), r.getString(2), r.getString(3),
			r.getFloat(4)));
    r.close();
  }

  public Account[] getAccounts() throws AccessViolation {
    Account[] a = new Account[l.size()];
    l.copyInto(a);
    return a;
  }

  public Account getAccount(int acctno) throws AccessViolation {
    Account a = l.getAccount(acctno);
    if (a == null) throw new AccessViolation(
      "Database error: Can't find account number " + acctno + "!");
    return a;
  }

  public Account newAccount(float openBal)
   throws AccessViolation, DataInvalid {
    int acctno;
    ResultSet r = null;
    try {
      // fix balance
      openBal = AccountServ.round(openBal);
    } catch (DataInvalid de) {
      de.explanation = "Opening balance invalid: " + de.explanation;
      throw de;
    }
    try {
      // get a new acctno
      r = psAcctNo.executeQuery();
      if (!r.next()) throw new AccessViolation(
        "Database error: Unable to allocate a new account number");
      acctno = r.getInt(1);
      // create record
      psNew.setInt(1, acctno);
      psNew.setFloat(2, openBal);
      psNew.executeUpdate();
    } catch (SQLException e) {
      throw new AccessViolation(
        "Database error on account creation: " + e.getMessage());
    } finally {
      try { r.close(); } catch (Exception e) {}
    }
    // add to list and return new Account
    AccountServ a = new AccountServ(this, acctno, openBal);
    l.addAccount(a);
    return a;
  }
}
